{******************************************************************}
{                                                                  }
{       TTaskScheduler wrapper object                              }
{       TaskSched.pas                                              }
{                                                                  }
{       Written by Christian Ruocco (10/10/2001)                   }
{                                                                  }
{  Do what you will with this but for god sakes don't claim you    }
{  wrote it! If you have any hassles or comments or bugs don't     }
{  hesitate to email me at xris@ndc.co.za. If that fails, try      }
{  buboo@mweb.co.za                                                }
{                                                                  }
{  This has been tested in Win 98/ME/2000                          }
{  Windows NT doesn't seem to have the COM interfaces required but }
{  perhaps a latest service pack will help.                        }
{******************************************************************}


unit TaskSched;

interface

Uses MsTask, Classes, SysUtils;

type
  DWORD = Cardinal;

  ETaskNeverRun = Exception;
  ETaskNotFound = Exception;
  ETaskSchedulerNotFound = Exception;

  TTaskScheduler = class;
  TTaskSchedulerTask = class;
  TTaskSchedulerTasks = class;
  TTaskSchedulerTaskTriggers = class;

  TTaskTriggerKind = (tskOnce, tskDaily, tskWeekly, tskMonthlyDate,
                      tskMonthlyDayOfWeek, tskOnIdle, tskAtSystemStart,
                      tskAtLogon);

  TTaskTriggerDayOfWeek = (tsdSunday, tsdMonday, tsdTuesday, tsdWednesday, tsdThursday, tsdFriday, tsdSaterday);
  TTaskTriggerDaysOfWeek = set of TTaskTriggerDayOfWeek;

  TTaskTriggerMonth = (tsmJanuary, tsmFebruary, tsmMarch, tsmApril, tsmMay, tsmJune, tsmJuly, tsmAugust, tsmSeptember, tsmOctober, tsmNovember, tsmDecember);
  TTaskTriggerMonths = set of TTaskTriggerMonth;

  TTaskTriggerWeek = (tswFirst, tswSecond, tswThird, tswFourth, tswFifth);

  TTaskTriggerDayOfMonth = (tdm1st,   tdm2nd, tdm3rd,   tdm4th,  tdm5th,  tdm6th,  tdm7th,
                            tdm8th,   tdm9th, tdm10th, tdm11th, tdm12th, tdm13th, tdm14th,
                            tdm15th, tdm16th, tdm17th, tdm18th, tdm19th, tdm20th, tdm21st,
                            tdm22nd, tdm23rd, tdm24th, tdm25th, tdm26th, tdm27th, tdm28th,
                            tdm29th, tdm30th, tdm31st, tdm32nd);

  TTaskTriggerDaysOfMonth = set of TTaskTriggerDayOfMonth;

  TTaskSchedulerTaskTrigger = class
  private
    FIndex : Integer;
    FTriggers : TTaskSchedulerTaskTriggers;
    FTaskTrig : TTaskTrigger;
    function GetStartTime : TDateTime;
    procedure SetStartTime(Value : TDateTime);
    function GetStopTime : TDateTime;
    procedure SetStopTime(Value : TDateTime);
    function GetDescription : String;
    function GetKind : TTaskTriggerKind;
    procedure SetKind(Value : TTaskTriggerKind);
    function GetDaysOfWeek : TTaskTriggerDaysOfWeek;
    procedure SetDaysOfWeek(Value : TTaskTriggerDaysOfWeek);
    function GetMonths : TTaskTriggerMonths;
    procedure SetMonths(Value : TTaskTriggerMonths);
    function GetWeek : TTaskTriggerWeek;
    procedure SetWeek(Value : TTaskTriggerWeek);
    function GetDaysInterval : Integer;
    procedure SetDaysInterval(Value : Integer);
    function GetWeeksInterval : Integer;
    procedure SetWeeksInterval(Value : Integer);
    function GetDaysOfMonth : TTaskTriggerDaysOfMonth;
    procedure SetDaysOfMonth(Value : TTaskTriggerDaysOfMonth);
    function GetHasEndDate : Boolean;
    procedure SetHasEndDate(Value : Boolean);
    function GetKillAtDurationEnd : Boolean;
    procedure SetKillAtDurationEnd(Value : Boolean);
    function GetEnabled : Boolean;
    procedure SetEnabled(Value : Boolean);
  public
    constructor Create(ATriggers : TTaskSchedulerTaskTriggers; AIndex : Integer);
    destructor Destroy; override;
    function GetDaysOfMonthStr : String;
    function GetDaysOfWeekStr : String;
    function GetMonthsStr : String;
    function GetWeekStr : String;
    procedure Sync;
    procedure Save;
    property Enabled : Boolean read GetEnabled write SetEnabled;
    property KillAtDurationEnd : Boolean read GetKillAtDurationEnd write SetKillAtDurationEnd;
    property HasEndDate : Boolean read GetHasEndDate write SetHasEndDate;
    property StartTime : TDateTime read GetStartTime write SetStartTime;
    property StopTime : TDateTime read GetStopTime write SetStopTime;
    property DaysOfMonth : TTaskTriggerDaysOfMonth read GetDaysOfMonth write SetDaysOfMonth;
    property DaysInterval : Integer read GetDaysInterval write SetDaysInterval;
    property Description : String read GetDescription;
    property Triggers : TTaskSchedulerTaskTriggers read FTriggers;
    property Kind : TTaskTriggerKind read GetKind write SetKind;
    property DaysOfWeek : TTaskTriggerDaysOfWeek read GetDaysOfWeek write SetDaysOfWeek;
    property Months : TTaskTriggerMonths read GetMonths write SetMonths;
    property Week : TTaskTriggerWeek read GetWeek write SetWeek;
    property WeeksInterval : Integer read GetWeeksInterval write SetWeeksInterval;
  end;

  TTaskSchedulerTaskTriggers = class
  private
    FList : TList;
    FTask : TTaskSchedulerTask;
    function GetItemByIndex(Index : Integer) : TTaskSchedulerTaskTrigger;
  protected
    procedure FreeTriggers;
  public
    constructor Create(ATask : TTaskSchedulerTask);
    destructor Destroy;
    function Count : Word;
    procedure Sync;
    function Add : TTaskSchedulerTaskTrigger;
    property Task : TTaskSchedulerTask read FTask;
    property Items[Index : Integer] : TTaskSchedulerTaskTrigger read GetItemByIndex; default;
  end;

  TTaskSchedulerTask = class
  private
    FTasks : TTaskSchedulerTasks;
    FTriggers : TTaskSchedulerTaskTriggers;
    FName : String;
    function GetAppName : String;
    function GetStatusStr : String;
    function GetStatusCode : HRESULT;

    procedure SetAppName(Value : String);
    procedure SetMaxRunTime(Milliseconds : DWORD);
    procedure SetParameters(Value : String);
    procedure SetPriority(Value : DWORD);
    procedure SetTaskFlags(Value: DWORD);
    procedure SetWorkingDirectory(Value : String);

    function GetMaxRunTime : DWORD;
    function GetParameters : String;
    function GetPriority : DWORD;
    function GetTaskFlags : DWORD;
    function GetWorkingDirectory : String;
    function GetEnabled : Boolean;
    procedure SetEnabled(Value : Boolean);
  protected
    function GetITask : ITask;
  public
    constructor Create(ATaskSchedulerTasks : TTaskSchedulerTasks; AName : String);
    destructor Destroy; override;
    procedure Edit;
    property Enabled : Boolean read GetEnabled write SetEnabled;
    function GetNextRunTime : TDateTime;
    function GetMostRecentRunTime : TDateTime;
    property AppName : String read GetAppName write SetAppName;
    property MaxRunTime : DWORD read GetMaxRunTime write SetMaxRunTime;
    property Name : String read FName;
    property Parameters : String read GetParameters write SetParameters;
    property Priority : DWORD read GetPriority write SetPriority;
    property TaskFlags : DWORD read GetTaskFlags write SetTaskFlags;
    property Tasks : TTaskSchedulerTasks read FTasks;
    property Triggers : TTaskSchedulerTaskTriggers read FTriggers;
    property Status : String read GetStatusStr;
    property StatusCode : HRESULT read GetStatusCode;
    property WorkingDirectory : String read GetWorkingDirectory write SetWorkingDirectory;
  end;

  TTaskSchedulerTasks = class
  private
    FList : TList;
    FScheduler : TTaskScheduler;
    function GetItem(Index : Variant) : TTaskSchedulerTask;
    function GetIndexByName(Name : String) : Integer;
    function GetItemByName(Name : String) : TTaskSchedulerTask;
    procedure FreeTasks;
  protected
    procedure ISave(Task : ITask);
  public
    constructor Create(AScheduler : TTaskScheduler);
    destructor Destroy; override;
    procedure Sync;
    procedure Clear;
    function Add(AName : String) : TTaskSchedulerTask;
    function Count : Integer;
    procedure Delete(Index : Variant);
    property Items[Index : Variant] : TTaskSchedulerTask read GetItem; default;
    property Scheduler : TTaskScheduler read FScheduler;
  end;

  TTaskSchedulerStatus = (tssUp, tssDown);

  TTaskScheduler = class
  private
    FTasks : TTaskSchedulerTasks;
    FAgent: ITaskScheduler;
    FStatus : TTaskSchedulerStatus;
  public
    constructor Create;
    destructor Destroy; override;
    property Tasks : TTaskSchedulerTasks read FTasks;
    property Agent : ITaskScheduler read FAgent;
    property Status : TTaskSchedulerStatus read FStatus;
  end;

const
  TaskTriggerKindStr : Array[TTaskTriggerKind] of String =
  ('Once', 'Daily', 'Weekly', 'Monthly date', 'Monthy day of week', 'On idle',
   'At system start', 'At logon');
  TaskTriggerDayOfWeekStr : Array[TTaskTriggerDayOfWeek] of String =
  ('Sunday', 'Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saterday');
  TaskTriggerMonthStr : Array[TTaskTriggerMonth] of String =
  ('January', 'February', 'March', 'April', 'May', 'June', 'July', 'August', 'September', 'October', 'November', 'December');
  TaskTriggerWeekStr : Array[TTaskTriggerWeek] of String =
  ('First', 'Second', 'Third', 'Fourth', 'Fifth');

var
  TaskScheduler : TTaskScheduler;


function TaskSchedulerDayToDaysOfMonth(Day : Byte) : TTaskTriggerDaysOfMonth;

implementation

Uses Windows, WinSvc, ActiveX, Dialogs;


var
  wsBuffer : PWideChar;


function TaskSchedulerDayToDaysOfMonth(Day : Byte) : TTaskTriggerDaysOfMonth;
begin
  Result := TTaskTriggerDaysOfMonth(1 shl (Day - 1));
end;


function StrToWide(p_sSource: string): LPCWSTR;
begin
  result := StringToWideChar(p_sSource, wsBuffer, Length(p_sSource) + 1);
end;

function GetLocalComputerName: string;
var
  Count: Cardinal;
begin
  Count := MAX_COMPUTERNAME_LENGTH + 1;
  SetLength(Result, Count); { set the length }
  if GetComputerName(PCHar(Result), Count) then SetLength(Result, Count) { set the exact length if it succeeded }
  else
  begin
    RaiseLastWin32Error; { raise exception if it failed }
  end;
end;

function StartScheduler: DWORD;
const
  SCHED_CLASS = 'SAGEWINDOWCLASS';
  SCHED_TITLE = 'SYSTEM AGENT COM WINDOW';
  SCHED_SERVICE_APP_NAME = 'mstask.exe';
  SCHED_SERVICE_NAME = 'Schedule';
  MAX_PATH: DWORD = 256;
var
  StartupInfo: TStartupInfo;
  ProcessInformation: TProcessInformation;
  szApp: PChar;
  pszPath: PChar;
  hsc: SC_HANDLE;
  hSchSvc: SC_HANDLE;
  SvcStatus: TServiceStatus;
  ServiceArgVectors: PChar;
begin
  // Determine what version of OS we are running on.
  if Win32Platform = VER_PLATFORM_WIN32_WINDOWS then
  begin
    // Start the Win95 version of TaskScheduler.
    if FindWindow(SCHED_CLASS, SCHED_TITLE) <> 0 then
      // It is already running.
      Result := ERROR_SUCCESS
    else
    begin
      //
      //  Execute the task scheduler process.
      //
      FillChar(StartupInfo, sizeof(StartupInfo), 0);
      StartupInfo.cb := sizeof(TStartupInfo);

      GetMem(szApp, MAX_PATH);
      try
        if SearchPath(nil, SCHED_SERVICE_APP_NAME, nil, MAX_PATH,
          szApp, pszPath) = 0 then
          Result := GetLastError
        else
        begin
          if not CreateProcess(szApp, nil, nil, nil, False,
            CREATE_NEW_CONSOLE or CREATE_NEW_PROCESS_GROUP, nil, nil,
            StartupInfo, ProcessInformation) then
            Result := GetLastError
          else
          begin
            CloseHandle(ProcessInformation.hProcess);
            CloseHandle(ProcessInformation.hThread);
            Result := ERROR_SUCCESS;
          end;
        end;
      finally
        FreeMem(szApp);
      end;
    end;
  end
  else
  begin
    // If not Win95 then start the NT version as a TaskScheduler service.

    hsc := OpenSCManager(nil, nil, SC_MANAGER_CONNECT);

    if hsc = 0 then
      Result := GetLastError
    else
    begin
      hSchSvc := OpenService(hsc, SCHED_SERVICE_NAME,
        SERVICE_START or SERVICE_QUERY_STATUS);

      CloseServiceHandle(hsc);

      if hSchSvc = 0 then
        Result := GetLastError
      else
      begin
        if not QueryServiceStatus(hSchSvc, SvcStatus) then
        begin
          CloseServiceHandle(hSchSvc);
          Result := GetLastError;
        end
        else
        begin
          if SvcStatus.dwCurrentState = SERVICE_RUNNING then
          begin
            // The service is already running.
            CloseServiceHandle(hSchSvc);
            Result := ERROR_SUCCESS;
          end
          else
          begin
            ServiceArgVectors := nil;
            if not StartService(hSchSvc, 0, ServiceArgVectors) then
            begin
              CloseServiceHandle(hSchSvc);
              Result := GetLastError();
            end
            else
            begin
              CloseServiceHandle(hSchSvc);
              Result := ERROR_SUCCESS;
            end;
          end;
        end;
      end;
    end;
  end;
end;

{ TTaskScheduler }

constructor TTaskScheduler.Create;
var
  Result : HRESULT;
begin
  FStatus := tssDown;

  if StartScheduler <> ERROR_SUCCESS then
    raise ETaskSchedulerNotFound.Create('Task Scheduler could not be loaded. Make sure it is installed correctly.');

  Result := ActiveX.CoInitialize(nil);

  if not (Result in [S_OK, S_FALSE]) then
    begin
      ActiveX.CoUninitialize;
      //ShowMessage('Error in ActiveX.CoInitialize.Error Code :' + IntToHex(Result, 8));
      Exit;
    end;

  Result := ActiveX.CoCreateInstance(CLSID_CSchedulingAgent,
    nil,
    CLSCTX_INPROC_SERVER,
    IID_ITaskScheduler,
    FAgent);

  if Result <> S_OK then
    begin
      ActiveX.CoUninitialize;
      //ShowMessage('Error in ActiveX.CoCreateInstance.Error Code :' + IntToHex(Result, 8));
      Exit;
    end;

  if FAgent.SetTargetComputer(StrToWide('\\' + GetLocalComputerName)) <> S_OK then
    showmessage('WHOOPS!');

  FTasks := TTaskSchedulerTasks.Create(Self);

  Tasks.Sync;

  FStatus := tssUp;
end;

destructor TTaskScheduler.Destroy;
begin
  FTasks.Free;
  FTasks := nil;
  FAgent := Nil;
  ActiveX.CoUninitialize;
  inherited;
end;


{ TTaskSchedulerTasks }

constructor TTaskSchedulerTasks.Create(AScheduler : TTaskScheduler);
begin
  FScheduler := AScheduler;
  FList := TList.Create;
  Sync;
end;

destructor TTaskSchedulerTasks.Destroy;
begin
  FreeTasks;
  FList.Free;
  FList := nil;
  FScheduler := Nil;
  inherited;
end;

function TTaskSchedulerTasks.Add(AName : String) : TTaskSchedulerTask;
var
  iiTask : ITask;
  WorkItem : IUnknown;
  hr : HRESULT;
  Task : TTaskSchedulerTask;
begin
  hr := Scheduler.Agent.NewWorkItem(StrToWide(AName), CLSID_CTask, IID_IScheduledWorkItem, WorkItem);

  iiTask := ITask(WorkItem);

  //ScheduledWorkItem := IScheduledWorkItem(WorkItem);

  if HR <> S_OK then
  begin
    raise Exception.Create(ClassName+'.Add(): Could not create the job ' + AName + ' (Error 0x' + IntTohex(hr, 8) + ')');
    exit;
  end;

  ISave(iiTask);

  Result := TTaskSchedulerTask.Create(Self, AName);
  FList.Add(Result);
end;

function TTaskSchedulerTasks.Count : Integer;
begin
  Result := FList.Count;
end;

function TTaskSchedulerTasks.GetIndexByName(Name : String) : Integer;
var
  I : Integer;
begin
  Name := UpperCase(Name);

  for I := 0 to Count - 1 do
    if UpperCase(TTaskSchedulerTask(FList.Items[I]).Name) = Name then
      begin
        Result := I;
        Exit;
      end;

  Result := -1;
end;


function TTaskSchedulerTasks.GetItemByName(Name : String) : TTaskSchedulerTask;
var
  I : Integer;
begin
  I := GetIndexByName(Name);
  if I = -1 then raise ETaskNotFound.Create('Task ''' + Name + ''' not found.');
  Result := FList.Items[I];
end;

function TTaskSchedulerTasks.GetItem(Index : Variant) : TTaskSchedulerTask;
var
  I : Integer;
  Temp : String;
begin
  case VarType(Index) and VarTypeMask of
    varString:
      Result := GetItemByName(Index);
    varInteger:
      Result := FList.Items[Index];
    else
      raise Exception.Create(ClassName + '.GetItem(): Unsupported index variant type.');
  end;
end;

procedure TTaskSchedulerTasks.Delete(Index : Variant);
var
  Task : TTaskSchedulerTask;
  iiTask : ITask;
  I : Integer;
begin
  Scheduler.Agent.Delete(StrToWide(Items[Index].Name));
  I := -1;

  case VarType(Index) and VarTypeMask of
    varString:
      I := GetIndexByName(Index);
    varInteger:
      I := Index;
    else
      raise Exception.Create(ClassName + '.Delete(): Unsupported index variant type.');
  end;

  Items[Index].Free;
  FList.Delete(I);
end;

procedure TTaskSchedulerTasks.Clear;
begin
  while Count > 0 do Delete(0);
end;

procedure TTaskSchedulerTasks.FreeTasks;
begin
  while FList.Count > 0 do
  begin
    TTaskSchedulerTask(FList.Items[0]).Free;
    FList.Delete(0);
  end;
end;

procedure TTaskSchedulerTasks.Sync;
var
  pEnum: IEnumWorkItems;
  hr : HRESULT;
  TaskNames: PLPWSTR;
  dwFetched: DWORD;
  Task : TTaskSchedulerTask;
begin
  hr := Scheduler.Agent.Enum(pEnum);

  if hr <> S_OK then raise Exception.Create(ClassName+'.Sync(): Couldn''t enumerate work items');

  FreeTasks;

  while (pEnum.Next(1, TaskNames, dwFetched) = 0) and (dwFetched > 0) do
  begin
    Task := TTaskSchedulerTask.Create(Self, WideCharToString(TaskNames^));
    FList.Add(Task);
    ActiveX.CoTaskMemFree(TaskNames);
  end;

end;

procedure TTaskSchedulerTasks.ISave(Task : ITask);
var
  pIPersistFile: IPersistFile;
  Result : DWORD;
begin
  Result := S_OK;
  Task.QueryInterface(IID_IPersistFile, pIPersistFile);

  if pIPersistFile <> nil then
    Result := pIPersistFile.Save(nil, true)
  else
    raise Exception.Create('Error in  Task.QueryInterface(IID_IPersistFile, pIPersistFile):');

  if Result <> S_OK then ShowMessage('.Save: err = 0x' + IntToHex(Result, 8));
end;



{ TTaskSchedulerTask }

constructor TTaskSchedulerTask.Create(ATaskSchedulerTasks : TTaskSchedulerTasks; AName : String);
begin
  FTasks := ATaskSchedulerTasks;
  FName := AName;
  FTriggers := TTaskSchedulerTaskTriggers.Create(Self);
end;

destructor TTaskSchedulerTask.Destroy;
begin
  FTriggers.Free;
  FTriggers := nil;
  FTasks := nil;
  inherited;
end;

procedure TTaskSchedulerTask.Edit;
var
  hr : HRESULT;
begin
  hr := GetITask.EditWorkItem(0, 0);
end;


function TTaskSchedulerTask.GetITask : ITask;
var
  PITask: IUnknown;
  hr: HRESULT;
  Temp : PWideChar;
begin
  Temp := StrToWide(Name);

  hr := Tasks.Scheduler.Agent.Activate(Temp, IID_ITask, PITask);

  if hr = S_OK then
    Result := ITask(PITask)
  else
    begin
      ShowMessage('Error in SchedulingAgent.Activate.Error Code :' + IntToHex(hr, 8));
      Result := nil;
    end;
end;

function TTaskSchedulerTask.GetAppName : String;
var
  Temp : PWideChar;
begin
  GetITask.GetApplicationName(Temp);
  Result := WideCharToString(Temp);

  ActiveX.CoTaskMemFree(Temp);
end;

function TTaskSchedulerTask.GetNextRunTime : TDateTime;
var
  Time : _SYSTEMTIME;
begin
  GetITask.GetNextRunTime(Time);
  //if hr <> S_OK then ListItem.SubItems.Add(DateTimeToStr(SystemTimeToDateTime(NextRunTime)))
  if Time.wYear = 0 then
    Result := 0
  else
    Result := SystemTimeToDateTime(Time);
end;

function TTaskSchedulerTask.GetMostRecentRunTime : TDateTime;
var
  Time : _SYSTEMTIME;
  hr : HRESULT;
begin
  hr := GetITask.GetMostRecentRunTime(Time);

  if hr = SCHED_S_TASK_HAS_NOT_RUN then
    raise ETaskNeverRun.Create('The task has never run.');

  Result := SystemTimeToDateTime(Time);
end;

function TTaskSchedulerTask.GetStatusCode : HRESULT;
begin
  GetITask.GetStatus(Result);
end;

function TTaskSchedulerTask.GetStatusStr : String;
begin
  case GetStatusCode of
    SCHED_S_TASK_READY:
      Result := 'The task is ready to run at its next scheduled time.';

    SCHED_S_TASK_RUNNING:
      Result := 'The task is currently running.';

    SCHED_S_TASK_DISABLED:
      Result := 'The task will not run at the scheduled times because it has been disabled.';

    SCHED_S_TASK_HAS_NOT_RUN:
      Result := 'The task has not yet run.';

    SCHED_S_TASK_NO_MORE_RUNS:
      Result := 'There are no more runs scheduled for this task.';

    SCHED_S_TASK_NOT_SCHEDULED:
      Result := 'One or more of the properties that are needed to run this task on a schedule have not been set.';

    SCHED_S_TASK_TERMINATED:
      Result := 'The last run of the task was terminated by the user.';

    SCHED_S_TASK_NO_VALID_TRIGGERS:
      Result := 'Either the task has no triggers or the existing triggers are disabled or not set.';

    SCHED_S_EVENT_TRIGGER:
      Result := 'Event triggers dont have set run times.';

    SCHED_E_TRIGGER_NOT_FOUND:
      Result := 'Trigger not found.';

    SCHED_E_TASK_NOT_READY:
      Result := 'One or more of the properties that are needed to run this task have not been set.';

    SCHED_E_TASK_NOT_RUNNING:
      Result := 'There is no running instance of the task to terminate.';

    SCHED_E_SERVICE_NOT_INSTALLED:
      Result := 'The Task Scheduler Service is not installed on this computer.';

    SCHED_E_CANNOT_OPEN_TASK:
      Result := 'The task object could not be opened.';

    SCHED_E_INVALID_TASK:
      Result := 'The object is either an invalid task object or is not a task object.';

    SCHED_E_ACCOUNT_INFORMATION_NOT_SET:
      Result := 'No account information could be found in the Task Scheduler security database for the task indicated.';

    SCHED_E_ACCOUNT_NAME_NOT_FOUND:
      Result := 'Unable to establish existence of the account specified.';

    SCHED_E_ACCOUNT_DBASE_CORRUPT:
      Result := 'Corruption was detected in the Task Scheduler security database; the database has been reset.';

    SCHED_E_NO_SECURITY_SERVICES:
      Result := 'Task Scheduler security services are available only on Windows NT.';


    SCHED_E_UNKNOWN_OBJECT_VERSION:
      Result := 'The task object version is either unsupported or invalid.';

    SCHED_E_UNSUPPORTED_ACCOUNT_OPTION:
      Result := 'The task has been configured with an unsupported combination of account settings and run time options.';

    SCHED_E_SERVICE_NOT_RUNNING:
      Result := 'The Task Scheduler Service is not running.';

    E_INVALIDARG:
      Result := 'The arguments are not valid.';

    E_OUTOFMEMORY:
      Result := 'Not enough memory is available.';
  end;

end;


procedure TTaskSchedulerTask.SetAppName(Value : String);
var
  hr : HRESULT;
  Task : ITask;
begin
  Task := GetITask;
  hr := Task.SetApplicationName(StrToWide(Value));
  if hr = S_OK then Tasks.ISave(Task);
end;

procedure TTaskSchedulerTask.SetMaxRunTime(Milliseconds : DWORD);
var
  hr : HRESULT;
  Task : ITask;
begin
  Task := GetITask;
  hr := Task.SetMaxRunTime(Milliseconds);
  if hr = S_OK then Tasks.ISave(Task);
end;

procedure TTaskSchedulerTask.SetParameters(Value : String);
var
  hr : HRESULT;
  Task : ITask;
begin
  Task := GetITask;
  hr := Task.SetParameters(StrToWide(Value));
  if hr = S_OK then Tasks.ISave(Task);
end;

procedure TTaskSchedulerTask.SetPriority(Value : DWORD);
var
  hr : HRESULT;
  Task : ITask;
begin
  Task := GetITask;
  Task.SetPriority(Value);
  if hr = S_OK then Tasks.ISave(Task);
end;

procedure TTaskSchedulerTask.SetTaskFlags(Value: DWORD);
var
  hr : HRESULT;
  Task : ITask;
begin
  Task := GetITask;
  hr := Task.SetTaskFlags(Value);
  if hr = S_OK then Tasks.ISave(Task);
end;

function TTaskSchedulerTask.GetMaxRunTime : DWORD;
begin
  GetITask.GetMaxRunTime(Result);
end;

function TTaskSchedulerTask.GetParameters : String;
var
  Temp : PWideChar;
begin
  GetITask.GetParameters(Temp);
  Result := WideCharToString(Temp);
  ActiveX.CoTaskMemFree(Temp);
end;

function TTaskSchedulerTask.GetPriority : DWORD;
begin
  GetITask.GetPriority(Result);
end;

function TTaskSchedulerTask.GetTaskFlags : DWORD;
begin
  GetITask.GetTaskFlags(Result);
end;

function TTaskSchedulerTask.GetWorkingDirectory : String;
var
  Temp : PWideChar;
begin
  GetITask.GetWorkingDirectory(Temp);
  Result := WideCharToString(Temp);
  ActiveX.CoTaskMemFree(Temp);
end;

procedure TTaskSchedulerTask.SetWorkingDirectory(Value : String);
var
  hr : HRESULT;
  Task : ITask;
begin
  Task := GetITask;
  hr := Task.SetWorkingDirectory(StrToWide(Value));
  if hr = S_OK then Tasks.ISave(Task);
end;

function TTaskSchedulerTask.GetEnabled : Boolean;
var
  Flags : DWORD;
begin
  GetITask.GetTaskFlags(Flags);
  Result := (Flags and TASK_FLAG_DISABLED) = 0;
end;

procedure TTaskSchedulerTask.SetEnabled(Value : Boolean);
var
  Flags : DWORD;
  iiTask : ITask;
begin
  iiTask := GetITask;

  iiTask.GetTaskFlags(Flags);

  if not Value then
    Flags := Flags or TASK_FLAG_DISABLED
  else
    Flags := Flags and (not TASK_FLAG_DISABLED);

  iiTask.SetTaskFlags(Flags);
  Tasks.ISave(iiTask);
end;


{ TTaskSchedulerTaskTrigger }

constructor TTaskSchedulerTaskTrigger.Create(ATriggers : TTaskSchedulerTaskTriggers; AIndex : Integer);
begin
  FIndex := AIndex;
  FTriggers := ATriggers;
  Sync;
end;

destructor TTaskSchedulerTaskTrigger.Destroy;
begin
  FTriggers.Free;
  FTriggers := nil;
end;

procedure TTaskSchedulerTaskTrigger.Save;
var
  iTrigger : ITaskTrigger;
  hr : HRESULT;
  iiTask : ITask;
begin
  iiTask := Triggers.Task.GetITask;
  iiTask.GetTrigger(FIndex, iTrigger);
  FTaskTrig.cbTriggerSize := sizeof(TASK_TRIGGER);
  hr := iTrigger.SetTrigger(@FTaskTrig);
  Triggers.Task.Tasks.ISave(iiTask);
end;

procedure TTaskSchedulerTaskTrigger.Sync;
var
  iTrigger : ITaskTrigger;
  hr : HRESULT;
  iiTask : ITask;
begin
  iiTask := Triggers.Task.GetITask;
  iiTask.GetTrigger(FIndex, iTrigger);
  FTaskTrig.cbTriggerSize := sizeof(TASK_TRIGGER);
  hr := iTrigger.GetTrigger(@FTaskTrig);
end;

function TTaskSchedulerTaskTrigger.GetStartTime : TDateTime;
begin
  with FTaskTrig do
    Result := EncodeDate(wBeginYear, wBeginMonth, wBeginDay) +
              EncodeTime(wStartHour, wStartMinute, 0, 0);
end;

procedure TTaskSchedulerTaskTrigger.SetStartTime(Value : TDateTime);
var
  Sec, MSec : Word;
begin
  Sec := 0;
  MSec := 0;
  with FTaskTrig do
  begin
    DecodeDate(Value, wBeginYear, wBeginMonth, wBeginDay);
    DecodeTime(Value, wStartHour, wStartMinute, Sec, MSec);
  end;
end;

function TTaskSchedulerTaskTrigger.GetStopTime : TDateTime;
begin
  with FTaskTrig do
    if (wEndYear or wEndMonth or wEndDay) <> 0 then
      Result := EncodeDate(wEndYear, wEndMonth, wEndDay)
    else
      Result := 0;
end;

procedure TTaskSchedulerTaskTrigger.SetStopTime(Value : TDateTime);
begin
  with FTaskTrig do DecodeDate(Value, wEndYear, wEndMonth, wEndDay);
end;

function TTaskSchedulerTaskTrigger.GetDescription : String;
var
  iTrigger : ITaskTrigger;
  iiTask : ITask;
  hr : HRESULT;
  Temp : PWideChar;
begin
  iiTask := Triggers.Task.GetITask;
  iiTask.GetTrigger(FIndex, iTrigger);
  hr := iTrigger.GetTriggerString(Temp);
  Result := WideCharToString(Temp);
end;


function TTaskSchedulerTaskTrigger.GetKind : TTaskTriggerKind;
begin
  Result := TTaskTriggerKind(FTaskTrig.TriggerType);
end;

procedure TTaskSchedulerTaskTrigger.SetKind(Value : TTaskTriggerKind);
begin
  FTaskTrig.TriggerType := _TASK_TRIGGER_TYPE(Value);
end;

function TTaskSchedulerTaskTrigger.GetDaysInterval : Integer;
begin
  if FTaskTrig.TriggerType <> TASK_TIME_TRIGGER_DAILY then
    raise Exception.Create(ClassName + '.GetDaysInterval(): Trigger kind must be daily to specify days interval.');

  Result := FTaskTrig.Type_.Daily.DaysInterval;
end;

procedure TTaskSchedulerTaskTrigger.SetDaysInterval(Value : Integer);
begin
  if FTaskTrig.TriggerType <> TASK_TIME_TRIGGER_DAILY then
    raise Exception.Create(ClassName + '.GetDaysInterval(): Trigger kind must be daily to specify days interval.');

  FTaskTrig.Type_.Daily.DaysInterval := Value;
end;


function TTaskSchedulerTaskTrigger.GetWeeksInterval : Integer;
begin
  if FTaskTrig.TriggerType <> TASK_TIME_TRIGGER_WEEKLY then
    raise Exception.Create(ClassName + '.GetWeeksInterval(): Trigger kind must be weelky to specify week interval.');

  Result := FTaskTrig.Type_.Weekly.WeeksInterval;
end;

procedure TTaskSchedulerTaskTrigger.SetWeeksInterval(Value : Integer);
begin
  if FTaskTrig.TriggerType <> TASK_TIME_TRIGGER_WEEKLY then
    raise Exception.Create(ClassName + '.SetWeeksInterval(): Trigger kind must be weelky to specify week interval.');

  FTaskTrig.Type_.Weekly.WeeksInterval := Value;
end;


function TTaskSchedulerTaskTrigger.GetDaysOfMonth : TTaskTriggerDaysOfMonth;
begin
  if FTaskTrig.TriggerType <> TASK_TIME_TRIGGER_MONTHLYDATE then
    raise Exception.Create(ClassName + '.GetDayOfMonth(): Trigger kind must be monthly date.');
  Result := TTaskTriggerDaysOfMonth(FTaskTrig.Type_.MonthlyDate.rgfDays);
end;

procedure TTaskSchedulerTaskTrigger.SetDaysOfMonth(Value : TTaskTriggerDaysOfMonth);
begin
   if FTaskTrig.TriggerType <> TASK_TIME_TRIGGER_MONTHLYDATE then
    raise Exception.Create(ClassName + '.GetDayOfMonth(): Trigger kind must be monthly date.');

  FTaskTrig.Type_.MonthlyDate.rgfDays := DWORD(Value)
end;


function TTaskSchedulerTaskTrigger.GetDaysOfMonthStr : String;
var
  N, X : Integer;
begin
  X := Cardinal(GetDaysOfMonth);

  for N := 0 to 31 do
  begin
    if X and 1 > 0 then
      begin
        Result := Result + IntToStr(N + 1);
        case (N + 1) mod 10 of
          1 : Result := Result + 'st';
          2 : Result := Result + 'nd';
          3 : Result := Result + 'rd';
        else
          Result := Result + 'th';
        end;
        Result := Result + ', ';
      end;
    X := X shr 1;
  end;

  if Length(Result) > 0 then SetLength(Result, Length(Result) - 2);
end;

function TTaskSchedulerTaskTrigger.GetDaysOfWeek : TTaskTriggerDaysOfWeek;
var
  Temp : ^Byte;
begin
  if not (FTaskTrig.TriggerType in [TASK_TIME_TRIGGER_WEEKLY, TASK_TIME_TRIGGER_MONTHLYDOW]) then
    raise Exception.Create(ClassName + '.GetDaysOfWeek(): Trigger kind must be weekly to specify days of week.');


  if FTaskTrig.TriggerType = TASK_TIME_TRIGGER_WEEKLY then
    Temp := @FTaskTrig.Type_.Weekly.rgfDaysOfTheWeek
  else
    Temp := @FTaskTrig.Type_.MonthlyDOW.rgfDaysOfTheWeek;

  Result := TTaskTriggerDaysOfWeek(Temp^);
end;

procedure TTaskSchedulerTaskTrigger.SetDaysOfWeek(Value : TTaskTriggerDaysOfWeek);
begin
  if not (FTaskTrig.TriggerType in [TASK_TIME_TRIGGER_WEEKLY, TASK_TIME_TRIGGER_MONTHLYDOW]) then
    raise Exception.Create(ClassName + '.GetDaysOfWeek(): Trigger kind must be weekly to specify days of week.');

  if FTaskTrig.TriggerType = TASK_TIME_TRIGGER_WEEKLY then
    FTaskTrig.Type_.Weekly.rgfDaysOfTheWeek := Byte(Value)
  else
    FTaskTrig.Type_.MonthlyDOW.rgfDaysOfTheWeek := Byte(Value);
end;

function TTaskSchedulerTaskTrigger.GetDaysOfWeekStr : String;
var
  N : TTaskTriggerDayOfWeek;
  Temp : Byte;
begin
  Temp := Byte(DaysOfWeek);

  if Temp <> $ff then
    begin
      for N := tsdSunday to tsdSaterday do
      begin
        if Temp and 1 <> 0 then Result := Result + TaskTriggerDayOfWeekStr[N] + ', ';
        Temp := Temp shr 1;
      end;

      if Length(Result) > 0 then SetLength(Result, Length(Result) - 2);
    end
  else
    Result := 'Every day';
end;

function TTaskSchedulerTaskTrigger.GetMonths : TTaskTriggerMonths;
begin
  if not (FTaskTrig.TriggerType in [TASK_TIME_TRIGGER_MONTHLYDOW,
                                   TASK_TIME_TRIGGER_MONTHLYDATE]) then
    raise Exception.Create(ClassName + '.GetMonths(): Trigger kind must be monthly day of week to specify months.');

  Result := TTaskTriggerMonths(FTaskTrig.Type_.MonthlyDate.rgfMonths);
end;

procedure TTaskSchedulerTaskTrigger.SetMonths(Value : TTaskTriggerMonths);
begin
  if not (FTaskTrig.TriggerType in [TASK_TIME_TRIGGER_MONTHLYDOW,
                                   TASK_TIME_TRIGGER_MONTHLYDATE]) then
    raise Exception.Create(ClassName + '.SetMonths(): Trigger kind must be monthly day of week to specify months.');

  FTaskTrig.Type_.MonthlyDate.rgfMonths := Word(Value);
end;

function TTaskSchedulerTaskTrigger.GetHasEndDate : Boolean;
begin
  Result := FTaskTrig.rgFlags and TASK_TRIGGER_FLAG_HAS_END_DATE > 0;
end;

procedure TTaskSchedulerTaskTrigger.SetHasEndDate(Value : Boolean);
begin
  if Value then
    FTaskTrig.rgFlags := FTaskTrig.rgFlags or TASK_TRIGGER_FLAG_HAS_END_DATE
  else
    FTaskTrig.rgFlags := FTaskTrig.rgFlags and (not TASK_TRIGGER_FLAG_HAS_END_DATE);
end;

function TTaskSchedulerTaskTrigger.GetKillAtDurationEnd : Boolean;
begin
  Result := FTaskTrig.rgFlags and TASK_TRIGGER_FLAG_KILL_AT_DURATION_END > 0;
end;

procedure TTaskSchedulerTaskTrigger.SetKillAtDurationEnd(Value : Boolean);
begin
  if Value then
    FTaskTrig.rgFlags := FTaskTrig.rgFlags or TASK_TRIGGER_FLAG_KILL_AT_DURATION_END
  else
    FTaskTrig.rgFlags := FTaskTrig.rgFlags and (not TASK_TRIGGER_FLAG_KILL_AT_DURATION_END);
end;


function TTaskSchedulerTaskTrigger.GetEnabled : Boolean;
begin
  Result := (FTaskTrig.rgFlags and TASK_TRIGGER_FLAG_DISABLED) = 0;
end;

procedure TTaskSchedulerTaskTrigger.SetEnabled(Value : Boolean);
begin
  if not Value then
    FTaskTrig.rgFlags := FTaskTrig.rgFlags or TASK_TRIGGER_FLAG_DISABLED
  else
    FTaskTrig.rgFlags := FTaskTrig.rgFlags and (not TASK_TRIGGER_FLAG_DISABLED);
end;





function TTaskSchedulerTaskTrigger.GetMonthsStr : String;
var
  N : TTaskTriggerMonth;
  Temp : Word;
begin
  Temp := Word(Months);

  if Temp = $0fff then
    begin
      Result := 'Every month';
      Exit;
    end;

  for N := tsmJanuary to tsmDecember do
  begin
    if Temp and 1 <> 0 then Result := Result + TaskTriggerMonthStr[N] + ', ';
    Temp := Temp shr 1;
  end;

  if Length(Result) > 0 then SetLength(Result, Length(Result) - 2);
end;


function TTaskSchedulerTaskTrigger.GetWeek : TTaskTriggerWeek;
begin
  if FTaskTrig.TriggerType <> TASK_TIME_TRIGGER_MONTHLYDOW then
    raise Exception.Create(ClassName + '.GetMonths(): Trigger kind must be monthly day of week to specify week number.');

  Result := TTaskTriggerWeek(FTaskTrig.Type_.MonthlyDOW.wWhichWeek - 1);
end;

procedure TTaskSchedulerTaskTrigger.SetWeek(Value : TTaskTriggerWeek);
begin
  FTaskTrig.Type_.MonthlyDOW.wWhichWeek := Byte(Value) + 1;
end;

function TTaskSchedulerTaskTrigger.GetWeekStr : String;
begin
  Result := TaskTriggerWeekStr[GetWeek] + ' ' + GetDaysOfWeekStr;
end;


{ TTaskSchedulerTaskTriggers }

constructor TTaskSchedulerTaskTriggers.Create(ATask : TTaskSchedulerTask);
begin
  FList := TList.Create;
  FTask := ATask;
  Sync;
end;

destructor TTaskSchedulerTaskTriggers.Destroy;
begin
  FreeTriggers;
  FList.Free;
  FList := nil;
  inherited;
end;

function TTaskSchedulerTaskTriggers.Add : TTaskSchedulerTaskTrigger;
var
  iiTask : ITask;
  iiTaskTrigger : ITaskTrigger;
  Trigger : TTaskSchedulerTaskTrigger;
  id : Word;
begin
  iiTask := Task.GetITask;
  iiTask.CreateTrigger(id, iiTaskTrigger);
  Task.Tasks.ISave(iiTask);
  Trigger := TTaskSchedulerTaskTrigger.Create(Self, id);
  FList.Add(Trigger);
  Result := Trigger;
end;

function TTaskSchedulerTaskTriggers.GetItemByIndex(Index : Integer) : TTaskSchedulerTaskTrigger;
begin
  Result := FList.Items[Index];
end;

function TTaskSchedulerTaskTriggers.Count : Word;
var
  hr : HRESULT;
begin
  hr := Task.GetITask.GetTriggerCount(Result);
end;

procedure TTaskSchedulerTaskTriggers.FreeTriggers;
begin
  while FList.Count > 0 do
  begin
    TTaskSchedulerTaskTrigger(FList.Items[0]).Free;
    FList.Delete(0);
  end;
end;

procedure TTaskSchedulerTaskTriggers.Sync;
var
  Trigger : TTaskSchedulerTaskTrigger;
  I : Integer;
  N : Word;
  iiTask : ITask;
begin
  FreeTriggers;

  iiTask := Task.GetITask;

  iiTask.GetTriggerCount(N);

  for I := 0 to N - 1 do
  begin
    Trigger := TTaskSchedulerTaskTrigger.Create(Self, I);
    FList.Add(Trigger);
  end;
end;


{ Helper functions }

initialization
  GetMem(wsBuffer, 1024);

finalization
  freeMem(wsBuffer);


end.
